from collections import namedtuple
from datetime import datetime, timezone

from th2_data_services.interfaces.utils.converter import ITimestampConverter

_DatetimeTuple = namedtuple("DatetimeTuple", "datetime mantissa")


class DatetimeStringConverter(ITimestampConverter[str]):
    """Converts datetime strings.

    If you request microseconds but your timestamp has nanoseconds, they will be just cut (not rounding).

    Expected timestamp format "yyyy-MM-ddTHH:mm:ss[.SSSSSSSSS]Z".
    """

    @classmethod
    def parse_timestamp(cls, datetime_string: str) -> (str, str):
        try:
            dt_tuple = _DatetimeTuple("", "")  # ('2022-03-05T23:56:44', '0Z')
            timestamp = datetime.strptime(datetime_string, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)
        except ValueError:
            dt_tuple = _DatetimeTuple(*datetime_string.rsplit("."))
            timestamp = datetime.strptime(dt_tuple.datetime, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc)

        mantissa_wo_z = dt_tuple.mantissa[:-1]
        nanoseconds = f"{mantissa_wo_z:0<9}"
        seconds = int(timestamp.timestamp())

        return seconds, nanoseconds



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

#LOG import logging
from http import HTTPStatus
from typing import List, Generator, Optional, Union

import requests
from requests import Response
from urllib3 import PoolManager, exceptions
from urllib.parse import quote

from th2_data_services.provider.interfaces.source_api import IHTTPProviderSourceAPI
from th2_data_services.decode_error_handler import UNICODE_REPLACE_HANDLER

#LOG logger = logging.getLogger("th2_data_services")
#LOG logger.setLevel(logging.DEBUG)


class HTTPProvider6API(IHTTPProviderSourceAPI):
    def __init__(
        self,
        url: str,
        chunk_length: int = 65536,
        decode_error_handler: str = UNICODE_REPLACE_HANDLER,
        char_enc: str = "utf-8",
    ):
        """HTTP Provider6 API.

        Args:
            url: HTTP data source url.
            chunk_length: How much of the content to read in one chunk.
            char_enc: Encoding for the byte stream.
            decode_error_handler: Registered decode error handler.
        """
        self._url = self.__normalize_url(url)
        self._char_enc = char_enc
        self._chunk_length = chunk_length
        self._decode_error_handler = decode_error_handler

    def __normalize_url(self, url):
        if url is None:
                return url

        pos = len(url) - 1
        while url[pos] == '/' and pos >= 0:
            pos -= 1

        return url[:pos + 1]

    def __encode_url(self, url: str) -> str:
        return quote(url.encode(), "/:&?=")

    def get_url_message_streams(self) -> str:
        """REST-API `messageStreams` call returns a list of message stream names."""
        return self.__encode_url(f"{self._url}/messageStreams")

    def get_url_find_event_by_id(self, event_id: str) -> str:
        """REST-API `event` call returns a single event with the specified id."""
        return self.__encode_url(f"{self._url}/event/{event_id}")

    def get_url_find_events_by_id(self, *ids) -> str:
        """REST-API `events` call returns a list of events with the specified ids.

        Note, at a time you can request no more eventSearchChunkSize.

        Deprecated, use `get_url_find_event_by_id` instead.
        """
        query = ""
        for id in ids:
            query += f"ids={id}&"
        return self.__encode_url(f"{self._url}/events?{query[:-1]}")

    def get_url_find_message_by_id(self, message_id: str) -> str:
        """REST-API `message` call returns a single message with the specified id."""
        return self.__encode_url(f"{self._url}/message/{message_id}")

    def get_url_messages_filters(self) -> str:
        """SSE-API `filters/sse-messages` call returns all names of sse message filters.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return self.__encode_url(f"{self._url}/filters/sse-messages")

    def get_url_events_filters(self) -> str:
        """SSE-API `/filters/sse-events` call returns all names of sse event filters.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return self.__encode_url(f"{self._url}/filters/sse-events")

    def get_url_message_filter_info(self, filter_name: str) -> str:
        """SSE-API `filters/sse-messages/{filter name}` call returns filter info.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return self.__encode_url(f"{self._url}/filters/sse-messages/{filter_name}")

    def get_url_event_filter_info(self, filter_name: str) -> str:
        """SSE-API `filters/sse-events/{filter_name}` call returns filter info.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return self.__encode_url(f"{self._url}/filters/sse-events/{filter_name}")

    def get_url_match_event_by_id(self, event_id: str, filters: str = "") -> str:
        """REST-API `match/event/{id}` call returns boolean value.

        Checks that event with the specified id is matched by filter.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return self.__encode_url(f"{self._url}/match/event/{event_id}{filters}")

    def get_url_match_message_by_id(self, message_id: str, filters: str = "") -> str:
        """REST-API `match/message/{id}` call returns boolean value.

        Checks that message with the specified id is matched by filter.

        https://github.com/th2-net/th2-rpt-data-provider#filters-api
        """
        return self.__encode_url(f"{self._url}/match/message/{message_id}{filters}")

    def get_url_search_sse_events(
        self,
        start_timestamp: int,
        end_timestamp: Optional[int] = None,
        parent_event: Optional[str] = None,
        resume_from_id: Optional[str] = None,
        search_direction: Optional[str] = "next",
        result_count_limit: Union[int, float] = None,
        keep_open: Optional[bool] = False,
        limit_for_parent: Union[int, float] = None,
        metadata_only: Optional[bool] = True,
        attached_messages: Optional[bool] = False,
        filters: Optional[str] = None,
    ) -> str:
        """REST-API `search/sse/events` call create a sse channel of event metadata that matches the filter.

        https://github.com/th2-net/th2-rpt-data-provider#sse-requests-api
        """
        kwargs = {
            "startTimestamp": start_timestamp,
            "endTimestamp": end_timestamp,
            "resumeFromId": resume_from_id,
            "parentEvent": parent_event,
            "searchDirection": search_direction,
            "resultCountLimit": result_count_limit,
            "limitForParent": limit_for_parent,
            "keepOpen": keep_open,
            "metadataOnly": metadata_only,
            "attachedMessages": attached_messages,
        }

        query = ""
        url = f"{self._url}/search/sse/events?"
        for k, v in kwargs.items():
            if v is None:
                continue
            else:
                query += f"&{k}={v}"
        url = f"{url}{query[1:]}"
        if filters is not None:
            url += filters
        return self.__encode_url(url)

    def get_url_search_sse_messages(
        self,
        start_timestamp: int,
        stream: List[str],
        end_timestamp: Optional[int] = None,
        resume_from_id: Optional[str] = None,
        search_direction: Optional[str] = "next",
        result_count_limit: Union[int, float] = None,
        keep_open: bool = False,
        message_id: Optional[List[str]] = None,
        attached_events: bool = False,
        lookup_limit_days: Union[int, float] = None,
        filters: Optional[str] = None,
    ) -> str:
        """REST-API `search/sse/messages` call create a sse channel of messages that matches the filter.

        https://github.com/th2-net/th2-rpt-data-provider#sse-requests-api
        """
        kwargs = {
            "startTimestamp": start_timestamp,
            "endTimestamp": end_timestamp,
            "resumeFromId": resume_from_id,
            "stream": stream,
            "searchDirection": search_direction,
            "resultCountLimit": result_count_limit,
            "keepOpen": keep_open,
            "messageId": message_id,
            "attachedEvents": attached_events,
            "lookupLimitDays": lookup_limit_days,
        }

        query = ""
        url = f"{self._url}/search/sse/messages?"
        for k, v in kwargs.items():
            if v is None:
                continue
            if k == "stream":
                for s in stream:
                    query += f"&{k}={s}"
            else:
                query += f"&{k}={v}"
        url = f"{url}{query[1:]}"
        if filters is not None:
            url += filters
        return self.__encode_url(url)

    def execute_sse_request(self, url: str) -> Generator[bytes, None, None]:
        """Create stream connection.

        Args:
            url: Url.

        Yields:
             str: Response stream data.
        """
        headers = {"Accept": "text/event-stream"}
        http = PoolManager()
        response = http.request(method="GET", url=url, headers=headers, preload_content=False)

        if response.status != HTTPStatus.OK:
            for s in HTTPStatus:
                if s == response.status:
                    raise exceptions.HTTPError(f"{s.value} {s.phrase} ({s.description})")
            raise exceptions.HTTPError(f"Http returned bad status: {response.status}")

        for chunk in response.stream(self._chunk_length):
            yield chunk

        response.release_conn()

    def execute_request(self, url: str) -> Response:
        """Sends a GET request to provider.

        Args:
            url: Url for a get request to rpt-data-provider.

        Returns:
            requests.Response: Response data.
        """
        return requests.get(url)



import os
import re
import subprocess
import sys
from typing import List


def get_version_by_pip(package_name: str):  # noqa: D103
    pip_run = subprocess.run([sys.executable, "-m", "pip", "show", package_name], capture_output=True)
    if pip_run.returncode != 0:
        return None

    output = pip_run.stdout.decode()
    matches = re.findall(r"Version: (.*)", output)
    return matches[0]


def get_package_version(package_name: str):  # noqa: D103
    version = get_version_by_pip(package_name).strip()
    # TODO: get version in case of conda
    return version


def verify_grpc_version(valid: List[str]):  # noqa: D103
    if os.environ.get("TH2_DS_SKIP_GRPC_VERIFY", "0") == "1":
        return
    version = get_package_version("th2_grpc_data_provider")

    if not version:
        raise SystemError(f"Package th2_grpc_data_provider not found")

    if version not in valid:
        raise SystemError(f"There is unsupported version of th2_grpc_data_provider for v5 provider api ({version})")



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from th2_data_services.provider.interfaces.filter import IProviderFilter
from typing import Sequence, Any, Union

from th2_grpc_data_provider.data_provider_pb2 import Filter as grpc_Filter, FilterName as grpc_FilterName
import google.protobuf.wrappers_pb2


class Provider6Filter(IProviderFilter):
    """General interface for Filters of Provider v6."""

    def __init__(
        self,
        name: str,
        values: Union[str, int, float, Sequence[Union[str, int, float]]],
        negative: bool = False,
        conjunct: bool = False,
    ):
        """Filter constructor.

        Args:
            name (str): Filter name.
            values (Union[str, int, float, Sequence[Union[str, int, float]]]): One string with filter value or list of filter values.
            negative (bool):  If true, will match events/messages that do not match those specified values.
                If false, will match the events/messages by their values. Defaults to false.
            conjunct (bool): If true, each of the specific filter values should be applied
                If false, at least one of the specific filter values must be applied.
        """
        self.name = name

        if isinstance(values, (list, tuple)):
            self.values = [str(v) for v in values]
        else:
            self.values = [str(values)]

        self.negative = negative
        self.conjunct = conjunct

    def __repr__(self):
        class_name = self.__class__.__name__
        return (
            f"{class_name}("
            f"name='{self.name}', "
            f"values={self.values}, "
            f"negative='{self.negative}', "
            f"conjunct='{self.conjunct}')"
        )

    def url(self) -> str:
        """Generates the filter part of the HTTP protocol API.

        For help use this readme:
        https://github.com/th2-net/th2-rpt-data-provider#filters-api.

        Returns:
            str: Generated filter.
        """
        return (
            f"&filters={self.name}"
            + "".join([f"&{self.name}-values={val}" for val in self.values])
            + f"&{self.name}-negative={self.negative}"
            + f"&{self.name}-conjunct={self.conjunct}"
        )

    def grpc(self) -> grpc_Filter:
        """Generates the grpc object of the GRPC protocol API."""
        return grpc_Filter(
            name=grpc_FilterName(name=self.name),
            negative=google.protobuf.wrappers_pb2.BoolValue(value=self.negative),
            value=self.values,
            conjunct=google.protobuf.wrappers_pb2.BoolValue(value=self.conjunct),
        )


class _Provider6FilterBase(Provider6Filter):
    FILTER_NAME = "FILTER_NAME"

    def __init__(self, values: Sequence[Any], negative: bool = False, conjunct: bool = False):
        super().__init__(self.FILTER_NAME, values, negative, conjunct)


class Provider6EventFilter(_Provider6FilterBase):
    """Base class for Event Filters of Provider v5."""


class Provider6MessageFilter(_Provider6FilterBase):
    """Base class for Message Filters of Provider v5."""



from th2_data_services.interfaces.utils.converter import ITimestampConverter


class Th2TimestampConverter(ITimestampConverter[dict]):
    """Converts Th2 timestamps.

    If you request microseconds but your timestamp has nanoseconds, they will be just cut (not rounding).

    Expected timestamp format {'epochSecond': 123, 'nano': 500}.
    """

    @classmethod
    def parse_timestamp(cls, timestamp: dict) -> (str, str):
        seconds, nanoseconds = timestamp["epochSecond"], f"{timestamp['nano']:0>9}"
        return seconds, nanoseconds



from typing import Generator

import orjson as json
from orjson import JSONDecodeError


class BufferedJSONProcessor:
    def __init__(self, buffer_limit: int = 250):
        """BufferedJSONProcessor constructor.

        Args:
            buffer_limit: By default 250. If limit is 0 buffer will not be used.
        """
        self.buffer = []
        self.buffer_limit = buffer_limit
        if buffer_limit == 0:
            self.decode = self._decode_without_buffer
        else:
            self.decode = self._decode_with_buffer

    def from_buffer(self) -> Generator:
        """Transforms JSON objects to dict objects.

        Returns:
            Generator[dict]
        """
        try:
            for i in json.loads("[" + ",".join(self.buffer) + "]"):
                yield i
        except JSONDecodeError as e:
            raise ValueError(f"json.decoder.JSONDecodeError: Invalid json received.\n" f"{e}\n" f"{self.buffer}")
        self.buffer = []

    def _decode_without_buffer(self, x: str) -> dict:
        """Decode JSON without buffer.

        Args:
            x: JSON Object

        Returns:
            dict
        """
        yield json.loads(x)

    def _decode_with_buffer(self, x: str) -> dict:
        """Decode JSON with buffer.

        Args:
            x: JSON Object

        Returns:
            dict
        """
        if len(self.buffer) < self.buffer_limit:
            self.buffer.append(x)
        else:
            yield from self.from_buffer()
            self.buffer = [x]

    def fin(self) -> Generator:
        """If buffer exists returns dicts from buffer.

        Returns:
            Generator[dict]
        """
        if self.buffer:
            yield from self.from_buffer()



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from th2_data_services.provider.v5.filters.filter import Provider5EventFilter


class TypeFilter(Provider5EventFilter):
    """Will match the events which type contains one of the given substrings."""

    FILTER_NAME = "type"


class NameFilter(Provider5EventFilter):
    """Will match the events which name contains one of the given substrings."""

    FILTER_NAME = "name"


class BodyFilter(Provider5EventFilter):
    """Will match the events which body contains one of the given substrings."""

    FILTER_NAME = "body"


class AttachedMessageIdFilter(Provider5EventFilter):
    """Filters the events that are linked to the specified message id."""

    FILTER_NAME = "attachedMessageId"


class _StatusFilter(Provider5EventFilter):
    FILTER_NAME = "status"

    def url(self) -> str:
        """Generates the filter part of the HTTP protocol API.

        For help use this readme:
        https://github.com/th2-net/th2-rpt-data-provider#filters-api.

        Returns:
            str: Generated filter.
        """
        return f"&filters={self.name}" + "".join([f"&{self.name}-values={val}" for val in self.values])


class PassedStatusFilter(_StatusFilter):
    """Will match the events which status equals passed."""

    def __init__(self):  # noqa: D107
        super().__init__("passed")


class FailedStatusFilter(_StatusFilter):
    """Will match the events which status equals failed."""

    def __init__(self):  # noqa: D107
        super().__init__("failed")



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Any


class IAdapter(ABC):
    """High level interface for Adapter."""

    @abstractmethod
    def handle(self, record: Any) -> Any:
        pass


class IMessageAdapter(IAdapter):
    """Interface of Adapter for messages."""

    @abstractmethod
    def handle(self, message: dict) -> Any:
        pass


class IEventAdapter(IAdapter):
    """Interface of Adapter for events."""

    @abstractmethod
    def handle(self, event: dict) -> Any:
        pass



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from th2_data_services.provider.v5.filters.filter import Provider5MessageFilter


class TypeFilter(Provider5MessageFilter):
    """Will match the messages by their full type name."""

    FILTER_NAME = "type"


class BodyBinaryFilter(Provider5MessageFilter):
    """Will match the messages by their binary body."""

    FILTER_NAME = "bodyBinary"


class BodyFilter(Provider5MessageFilter):
    """Will match the messages by their parsed body."""

    FILTER_NAME = "body"


class AttachedEventIdsFilter(Provider5MessageFilter):
    """Filters the messages that are linked to the specified event id."""

    FILTER_NAME = "attachedEventIds"



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import warnings
from typing import Generator, Callable

from sseclient import Event as SSEEvent
from urllib3.exceptions import HTTPError
import orjson as json

from th2_data_services.interfaces import IAdapter
from th2_data_services.utils.json import BufferedJSONProcessor


class SSEAdapter(IAdapter):
    """SSE Adapter handles bytes from sse-stream into Dict object."""

    def __init__(self):
        warnings.warn("This class is deprecated please use StreamingSSEAdapter")

    def handle(self, record: SSEEvent) -> dict:
        """Adapter handler.

        Args:
            record: SSE Event.

        Returns:
            Dict object.
        """
        if record.event == "error":
            raise HTTPError(record.data)
        if record.event not in ["close", "keep_alive", "message_ids"]:
            try:
                return json.loads(record.data)
            except json.JSONDecodeError as e:
                raise ValueError(f"json.decoder.JSONDecodeError: Invalid json received.\n" f"{e}\n" f"{record.data}")


class StreamingSSEAdapter(IAdapter):
    def __init__(self, json_processor: BufferedJSONProcessor):
        self.json_processor = json_processor
        self.events_types_blacklist = {"close", "keep_alive", "message_ids"}

    def handle(self, record: (Generator[SSEEvent, None, None], Callable)) -> Generator[dict, None, None]:
        if callable(record):
            stream = record()
        else:
            stream = record
        for event in stream:
            if event.event == "error":
                raise HTTPError(event.data)
            if event.event not in self.events_types_blacklist:
                yield from self.json_processor.decode(event.data)
        yield from self.json_processor.fin()


def get_default_sse_adapter():
    return StreamingSSEAdapter(BufferedJSONProcessor())



#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from importlib_metadata import version
from typing import Sequence, Union
from warnings import warn

v = version("th2_grpc_data_provider")

if v == "1.1.0":  # v6
    from th2_data_services.provider.v6.filters.filter import Provider6Filter as ProviderFilter  # noqa
elif v == "0.1.6":  # v5
    from th2_data_services.provider.v5.filters.filter import Provider5Filter as ProviderFilter  # noqa


class Filter(ProviderFilter):
    """The class for using rpt-data-provider filters API."""

    def __init__(
        self,
        name: str,
        values: Union[str, int, float, Sequence[Union[str, int, float]]],
        negative: bool = False,
        conjunct: bool = False,
    ):
        """Filter constructor.

        Args:
            name (str): Filter name.
            values (Union[str, int, float, Sequence[Union[str, int, float]]]): One string with filter value or list of filter values.
            negative (bool):  If true, will match events/messages that do not match those specified values.
                If false, will match the events/messages by their values. Defaults to false.
            conjunct (bool): If true, each of the specific filter values should be applied
                If false, at least one of the specific filter values must be applied.
        """
        warn(
            f"{self.__class__.__name__} is deprecated. Use Filters of certain DataSource instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(name=name, values=values, negative=negative, conjunct=conjunct)



from typing import List

from th2_grpc_common.common_pb2 import Direction
from th2_grpc_data_provider.data_provider_pb2 import MessageStream


class Streams:
    """General interface for composite streams of Provider v6.

    The class gives the opportunity to make list of streams with direction for each.
    """

    def __init__(self, streams: List[str], direction: str = None):
        """Streams constructor.

        Args:
            streams: List of Streams.
            direction: Direction of Streams (Only FIRST or SECOND). If None then is both directions.
        """
        self._streams = streams
        if direction is not None:
            direction = direction.upper()
            if direction not in ("FIRST", "SECOND"):
                raise ValueError("The direction must be 'FIRST' or 'SECOND'.")
        self._direction = direction

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}(" f"streams={self._streams}, " f"direction={self._direction})"

    def url(self) -> str:
        """Generates the stream part of the HTTP protocol API.

        Returns:
            str: Generated streams.
        """
        if self._direction is None:
            return "&".join([f"stream={stream}:FIRST&stream={stream}:SECOND" for stream in self._streams])
        return "&".join([f"stream={stream}:{self._direction}" for stream in self._streams])

    def grpc(self) -> List[MessageStream]:
        """Generates the grpc objects of the GRPC protocol API.

        Returns:
            List[MessageStream]: List of Stream with specified direction.
        """
        if self._direction is None:
            result = []
            for stream in self._streams:
                result += [
                    MessageStream(name=stream, direction=Direction.Value("FIRST")),
                    MessageStream(name=stream, direction=Direction.Value("SECOND")),
                ]
            return result
        return [MessageStream(name=stream, direction=Direction.Value(self._direction)) for stream in self._streams]
